@@ -130,20 +130,21 @@ struct fuse_forget_data {
 struct fuse_custom_io {
 	ssize_t (*writev)(int fd, struct iovec *iov, int count, void *userdata);
 	ssize_t (*read)(int fd, void *buf, size_t buf_len, void *userdata);
-	ssize_t (*splice_receive)(int fdin, __off64_t *offin, int fdout,
-					  __off64_t *offout, size_t len,
+	ssize_t (*splice_receive)(int fdin, off_t *offin, int fdout,
+					  off_t *offout, size_t len,
 				  	  unsigned int flags, void *userdata);
-	ssize_t (*splice_send)(int fdin, __off64_t *offin, int fdout,
-				     __off64_t *offout, size_t len,
+	ssize_t (*splice_send)(int fdin, off_t *offin, int fdout,
+				     off_t *offout, size_t len,
 			           unsigned int flags, void *userdata);
 };
 
 /**
- * Flags for fuse_lowlevel_notify_expire_entry()
+ * Flags for fuse_lowlevel_notify_entry()
  * 0 = invalidate entry
  * FUSE_LL_EXPIRE_ONLY = expire entry
 */
-enum fuse_expire_flags {
+enum fuse_notify_entry_flags {
+	FUSE_LL_INVALIDATE = 0,
 	FUSE_LL_EXPIRE_ONLY	= (1 << 0),
 };
 
@@ -156,7 +157,15 @@ enum fuse_expire_flags {
 #define FUSE_SET_ATTR_MTIME	(1 << 5)
 #define FUSE_SET_ATTR_ATIME_NOW	(1 << 7)
 #define FUSE_SET_ATTR_MTIME_NOW	(1 << 8)
+#define FUSE_SET_ATTR_FORCE	(1 << 9)
 #define FUSE_SET_ATTR_CTIME	(1 << 10)
+#define FUSE_SET_ATTR_KILL_SUID	(1 << 11)
+#define FUSE_SET_ATTR_KILL_SGID	(1 << 12)
+#define FUSE_SET_ATTR_FILE	(1 << 13)
+#define FUSE_SET_ATTR_KILL_PRIV	(1 << 14)
+#define FUSE_SET_ATTR_OPEN	(1 << 15)
+#define FUSE_SET_ATTR_TIMES_SET	(1 << 16)
+#define FUSE_SET_ATTR_TOUCH	(1 << 17)
 
 /* ----------------------------------------------------------- *
  * Request methods and replies				       *
@@ -305,6 +314,12 @@ struct fuse_lowlevel_ops {
 	 * expected to reset the setuid and setgid bits if the file
 	 * size or owner is being changed.
 	 *
+	 * This method will not be called to update st_atime or st_ctime implicitly
+	 * (eg. after a read() request), and only be called to implicitly update st_mtime
+	 * if writeback caching is active. It is the filesystem's responsibility to update
+	 * these timestamps when needed, and (if desired) to implement mount options like
+	 * `noatime` or `relatime`.
+	 *
 	 * If the setattr was invoked from the ftruncate() system call
 	 * under Linux kernel versions 2.6.15 or later, the fi->fh will
 	 * contain the value set by the open method or will be undefined
@@ -1668,11 +1683,10 @@ int fuse_lowlevel_notify_inval_inode(struct fuse_session *se, fuse_ino_t ino,
 				     off_t off, off_t len);
 
 /**
- * Notify to invalidate parent attributes and the dentry matching
- * parent/name
+ * Notify to invalidate parent attributes and the dentry matching parent/name
  *
  * To avoid a deadlock this function must not be called in the
- * execution path of a related filesytem operation or within any code
+ * execution path of a related filesystem operation or within any code
  * that could hold a lock that could be needed to execute such an
  * operation. As of kernel 4.18, a "related operation" is a lookup(),
  * symlink(), mknod(), mkdir(), unlink(), rename(), link() or create()
@@ -1696,14 +1710,13 @@ int fuse_lowlevel_notify_inval_entry(struct fuse_session *se, fuse_ino_t parent,
 				     const char *name, size_t namelen);
 
 /**
- * Notify to expire or invalidate parent attributes and the dentry 
- * matching parent/name
+ * Notify to expire parent attributes and the dentry matching parent/name
  * 
- * Underlying function for fuse_lowlevel_notify_inval_entry().
+ * Same restrictions apply as for fuse_lowlevel_notify_inval_entry()
  * 
- * In addition to invalidating an entry, it also allows to expire an entry.
- * In that case, the entry is not forcefully removed from kernel cache 
- * but instead the next access to it forces a lookup from the filesystem.
+ * Compared to invalidating an entry, expiring the entry results not in a
+ * forceful removal of that entry from kernel cache but instead the next access
+ * to it forces a lookup from the filesystem.
  * 
  * This makes a difference for overmounted dentries, where plain invalidation
  * would detach all submounts before dropping the dentry from the cache. 
@@ -1714,17 +1727,18 @@ int fuse_lowlevel_notify_inval_entry(struct fuse_session *se, fuse_ino_t parent,
  * so invalidation will only be triggered for the non-overmounted case.
  * The dentry could also be mounted in a different mount instance, in which case
  * any submounts will still be detached.
+ * 
+ * Added in FUSE protocol version 7.38. If the kernel does not support
+ * this (or a newer) version, the function will return -ENOSYS and do nothing.
  *
  * @param se the session object
  * @param parent inode number
  * @param name file name
  * @param namelen strlen() of file name
- * @param flags flags to control if the entry should be expired or invalidated
- * @return zero for success, -errno for failure
+ * @return zero for success, -errno for failure, -enosys if no kernel support
 */
 int fuse_lowlevel_notify_expire_entry(struct fuse_session *se, fuse_ino_t parent,
-                                      const char *name, size_t namelen,
-                                      enum fuse_expire_flags flags);
+                                      const char *name, size_t namelen);
 
 /**
  * This function behaves like fuse_lowlevel_notify_inval_entry() with
@@ -1736,7 +1750,7 @@ int fuse_lowlevel_notify_expire_entry(struct fuse_session *se, fuse_ino_t parent
  * that the dentry has been deleted.
  *
  * To avoid a deadlock this function must not be called while
- * executing a related filesytem operation or while holding a lock
+ * executing a related filesystem operation or while holding a lock
  * that could be needed to execute such an operation (see the
  * description of fuse_lowlevel_notify_inval_entry() for more
  * details).
@@ -1958,7 +1972,7 @@ struct fuse_cmdline_opts {
  * @param opts output argument for parsed options
  * @return 0 on success, -1 on failure
  */
-#if (defined(HAVE_LIBC_VERSIONED_SYMBOLS))
+#if (defined(LIBFUSE_BUILT_WITH_VERSIONED_SYMBOLS))
 int fuse_parse_cmdline(struct fuse_args *args,
 		       struct fuse_cmdline_opts *opts);
 #else
@@ -2076,7 +2090,7 @@ int fuse_session_loop(struct fuse_session *se);
 	int fuse_session_loop_mt_32(struct fuse_session *se, struct fuse_loop_config *config);
 	#define fuse_session_loop_mt(se, config) fuse_session_loop_mt_32(se, config)
 #else
-	#if (defined(HAVE_LIBC_VERSIONED_SYMBOLS))
+	#if (defined(LIBFUSE_BUILT_WITH_VERSIONED_SYMBOLS))
 		/**
 		 * Enter a multi-threaded event loop.
 		 *
@@ -2098,9 +2112,12 @@ int fuse_session_loop(struct fuse_session *se);
 /**
  * Flag a session as terminated.
  *
- * This function is invoked by the POSIX signal handlers, when
- * registered using fuse_set_signal_handlers(). It will cause any
- * running event loops to terminate on the next opportunity.
+ * This will cause any running event loops to terminate on the next opportunity. If this function is
+ * called by a thread that is not a FUSE worker thread, the next
+ * opportunity will be when FUSE a request is received (which may be far in the future if the
+ * filesystem is not currently being used by any clients). One way to avoid this delay is to
+ * afterwards sent a signal to the main thread (if fuse_set_signal_handlers() is used, SIGPIPE
+ * will cause the main thread to wake-up but otherwise be ignored).
  *
  * @param se the session
  */
